@@ -1,12 +1,12 @@
 ;;; diff-hl.el --- Highlight uncommitted changes using VC -*- lexical-binding: t -*-
 
-;; Copyright (C) 2012-2020  Free Software Foundation, Inc.
+;; Copyright (C) 2012-2021  Free Software Foundation, Inc.
 
 ;; Author:   Dmitry Gutov <dgutov@yandex.ru>
 ;; URL:      https://github.com/dgutov/diff-hl
 ;; Keywords: vc, diff
-;; Version:  1.8.8
-;; Package-Requires: ((cl-lib "0.2") (emacs "24.3"))
+;; Version:  1.9.0
+;; Package-Requires: ((cl-lib "0.2") (emacs "25.1"))
 
 ;; This file is part of GNU Emacs.
 
@@ -21,7 +21,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -35,8 +35,15 @@
 ;; `diff-hl-revert-hunk'     C-x v n
 ;; `diff-hl-previous-hunk'   C-x v [
 ;; `diff-hl-next-hunk'       C-x v ]
+;; `diff-hl-show-hunk'       C-x v *
+;; `diff-hl-stage-current-hunk' C-x v S
+;; `diff-hl-set-reference-rev'
+;; `diff-hl-reset-reference-rev'
+;; `diff-hl-unstage-file'
 ;;
 ;; The mode takes advantage of `smartrep' if it is installed.
+;;
+;; Alternatively, it integrates with `repeat-mode' (Emacs 28+).
 
 ;; Add either of the following to your init file.
 ;;
@@ -55,6 +62,8 @@
 (require 'diff-mode)
 (require 'vc)
 (require 'vc-dir)
+(require 'log-view)
+
 (eval-when-compile
   (require 'cl-lib)
   (require 'vc-git)
@@ -97,6 +106,11 @@
   :group 'diff-hl
   :type 'boolean)
 
+(defcustom diff-hl-disable-on-remote nil
+  "Non-nil will disable `diff-hl' in remote buffers."
+  :group 'diff-hl
+  :type 'boolean)
+
 (defcustom diff-hl-ask-before-revert-hunk t
   "Non-nil to ask for confirmation before revert a hunk."
   :group 'diff-hl
@@ -126,6 +140,7 @@
   "Which side to use for indicators."
   :type '(choice (const left)
                  (const right))
+  :initialize 'custom-initialize-default
   :set (lambda (var value)
          (let ((on (bound-and-true-p global-diff-hl-mode)))
            (when on (global-diff-hl-mode -1))
@@ -133,13 +148,15 @@
            (when on (global-diff-hl-mode 1)))))
 
 (defcustom diff-hl-highlight-revert-hunk-function
-  #'diff-hl-revert-highlight-first-column
-  "Function to highlight the current hunk in `diff-hl-revert-hunk'.
-The function is called at the beginning of the hunk and passed
+  #'diff-hl-revert-narrow-to-hunk
+  "Function to emphasize the current hunk in `diff-hl-revert-hunk'.
+The function is called at the beginning of the hunk and is passed
 the end position as its only argument."
   :type '(choice (const :tag "Do nothing" ignore)
                  (const :tag "Highlight the first column"
-                        diff-hl-revert-highlight-first-column)))
+                        diff-hl-revert-highlight-first-column)
+                 (const :tag "Narrow to the hunk"
+                        diff-hl-revert-narrow-to-hunk)))
 
 (defcustom diff-hl-global-modes '(not image-mode)
   "Modes for which `diff-hl-mode' is automagically turned on.
@@ -161,6 +178,22 @@ performance when viewing such files in certain conditions."
                       (repeat :inline t (symbol :tag "mode"))))
   :group 'diff-hl)
 
+(defcustom diff-hl-show-staged-changes t
+  "Whether to include staged changes in the indicators.
+Only affects Git, it's the only backend that has staging area."
+  :type 'boolean)
+
+(defcustom diff-hl-goto-hunk-old-revisions nil
+  "When non-nil, `diff-hl-diff-goto-hunk' will always try to
+navigate to the line in the diff that corresponds to the current
+line in the file buffer (or as close as it can get to it).
+
+When this variable is nil (default), `diff-hl-diff-goto-hunk'
+only does that when called without the prefix argument, or when
+the NEW revision is not specified (meaning, the diff is against
+the current version of the file)."
+  :type 'boolean)
+
 (defvar diff-hl-reference-revision nil
   "Revision to diff against.  nil means the most recent one.")
 
@@ -246,6 +279,7 @@ performance when viewing such files in certain conditions."
     (t (intern (format "diff-hl-bmp-%s" type)))))
 
 (defvar vc-svn-diff-switches)
+(defvar vc-fossil-diff-switches)
 
 (defmacro diff-hl-with-diff-switches (body)
   `(let ((vc-git-diff-switches
@@ -260,6 +294,7 @@ performance when viewing such files in certain conditions."
                       vc-git-diff-switches))))
          (vc-hg-diff-switches nil)
          (vc-svn-diff-switches nil)
+         (vc-fossil-diff-switches '("-c" "0"))
          (vc-diff-switches '("-U0"))
          ,@(when (boundp 'vc-disable-async-diff)
              '((vc-disable-async-diff t))))
@@ -273,23 +308,32 @@ performance when viewing such files in certain conditions."
                ;; Diffing against an older revision.
                diff-hl-reference-revision))))
 
+(declare-function vc-git-command "vc-git")
+
 (defun diff-hl-changes-buffer (file backend)
-  ;; FIXME: To diff against the staging area, call 'git diff-files -p'.
-  (let ((buf-name " *diff-hl* "))
+  (diff-hl-with-diff-switches
+   (diff-hl-diff-against-reference file backend " *diff-hl* ")))
+
+(defun diff-hl-diff-against-reference (file backend buffer)
+  (if (and (eq backend 'Git)
+           (not diff-hl-reference-revision)
+           (not diff-hl-show-staged-changes))
+      (apply #'vc-git-command buffer 1
+             (list file)
+             "diff-files"
+             (cons "-p" (vc-switches 'git 'diff)))
     (condition-case err
-        (diff-hl-with-diff-switches
-         (vc-call-backend backend 'diff (list file)
-                          diff-hl-reference-revision nil
-                          buf-name))
+        (vc-call-backend backend 'diff (list file)
+                         diff-hl-reference-revision nil
+                         buffer)
       (error
        ;; https://github.com/dgutov/diff-hl/issues/117
        (when (string-match-p "\\`Failed (status 128)" (error-message-string err))
-         (diff-hl-with-diff-switches
-          (vc-call-backend backend 'diff (list file)
-                           "4b825dc642cb6eb9a060e54bf8d69288fbee4904"
-                           nil
-                           buf-name)))))
-    buf-name))
+         (vc-call-backend backend 'diff (list file)
+                          "4b825dc642cb6eb9a060e54bf8d69288fbee4904"
+                          nil
+                          buffer)))))
+  buffer)
 
 (defun diff-hl-changes ()
   (let* ((file buffer-file-name)
@@ -298,33 +342,47 @@ performance when viewing such files in certain conditions."
       (let ((state (vc-state file backend)))
         (cond
          ((diff-hl-modified-p state)
-          (let* (diff-auto-refine-mode res)
-            (with-current-buffer (diff-hl-changes-buffer file backend)
-              (goto-char (point-min))
-              (unless (eobp)
-                (ignore-errors
-                  (diff-beginning-of-hunk t))
-                (while (looking-at diff-hunk-header-re-unified)
-                  (let ((line (string-to-number (match-string 3)))
-                        (len (let ((m (match-string 4)))
-                               (if m (string-to-number m) 1)))
-                        (beg (point)))
-                    (diff-end-of-hunk)
-                    (let* ((inserts (diff-count-matches "^\\+" beg (point)))
-                           (deletes (diff-count-matches "^-" beg (point)))
-                           (type (cond ((zerop deletes) 'insert)
-                                       ((zerop inserts) 'delete)
-                                       (t 'change))))
-                      (when (eq type 'delete)
-                        (setq len 1)
-                        (cl-incf line))
-                      (push (list line len type) res))))))
-            (nreverse res)))
+          (diff-hl-changes-from-buffer
+           (diff-hl-changes-buffer file backend)))
          ((eq state 'added)
           `((1 ,(line-number-at-pos (point-max)) insert)))
          ((eq state 'removed)
           `((1 ,(line-number-at-pos (point-max)) delete))))))))
 
+(defun diff-hl-changes-from-buffer (buf)
+  (with-current-buffer buf
+    (let (res)
+      (goto-char (point-min))
+      (unless (eobp)
+        ;; TODO: When 27.1 is the minimum requirement, we can drop
+        ;; these bindings: that version, in addition to switching over
+        ;; to the diff-refine var, also added the
+        ;; called-interactively-p check, so refinement can't be
+        ;; triggered by code calling the navigation functions, only by
+        ;; direct interactive invocations.
+        (ignore-errors
+          (with-no-warnings
+            (let (diff-auto-refine-mode)
+              (diff-beginning-of-hunk t))))
+        (while (looking-at diff-hunk-header-re-unified)
+          (let ((line (string-to-number (match-string 3)))
+                (len (let ((m (match-string 4)))
+                       (if m (string-to-number m) 1)))
+                (beg (point)))
+            (with-no-warnings
+              (let (diff-auto-refine-mode)
+                (diff-end-of-hunk)))
+            (let* ((inserts (diff-count-matches "^\\+" beg (point)))
+                   (deletes (diff-count-matches "^-" beg (point)))
+                   (type (cond ((zerop deletes) 'insert)
+                               ((zerop inserts) 'delete)
+                               (t 'change))))
+              (when (eq type 'delete)
+                (setq len 1)
+                (cl-incf line))
+              (push (list line len type) res)))))
+      (nreverse res))))
+
 (defun diff-hl-update ()
   (let ((changes (diff-hl-changes))
         (current-line 1))
@@ -354,6 +412,7 @@ performance when viewing such files in certain conditions."
                     (hook '(diff-hl-overlay-modified)))
                 (overlay-put h 'diff-hl t)
                 (overlay-put h 'diff-hl-hunk t)
+                (overlay-put h 'diff-hl-hunk-type type)
                 (overlay-put h 'modification-hooks hook)
                 (overlay-put h 'insert-in-front-hooks hook)
                 (overlay-put h 'insert-behind-hooks hook)))))))))
@@ -410,25 +469,63 @@ performance when viewing such files in certain conditions."
   (when revert-buffer-preserve-modes
     (diff-hl-update)))
 
-(defun diff-hl-diff-goto-hunk-1 ()
+(defun diff-hl-diff-goto-hunk-1 (historic)
+  (defvar vc-sentinel-movepoint)
   (vc-buffer-sync)
   (let* ((line (line-number-at-pos))
-         (buffer (current-buffer)))
-    (vc-diff-internal t (vc-deduce-fileset) diff-hl-reference-revision nil t)
-    (vc-exec-after `(if (< (line-number-at-pos (point-max)) 3)
-                        (with-current-buffer ,buffer (diff-hl-remove-overlays))
-                      (diff-hl-diff-skip-to ,line)
+         (buffer (current-buffer))
+         (rev1 diff-hl-reference-revision)
+         rev2)
+    (when historic
+      (let ((revs (diff-hl-diff-read-revisions rev1)))
+        (setq rev1 (car revs)
+              rev2 (cdr revs))))
+    (vc-diff-internal t (vc-deduce-fileset) rev1 rev2 t)
+    (vc-run-delayed (if (< (line-number-at-pos (point-max)) 3)
+                        (with-current-buffer buffer (diff-hl-remove-overlays))
+                      (when (or (not rev2) diff-hl-goto-hunk-old-revisions)
+                        (diff-hl-diff-skip-to line))
                       (setq vc-sentinel-movepoint (point))))))
 
-(defun diff-hl-diff-goto-hunk ()
+(defun diff-hl-diff-goto-hunk (&optional historic)
   "Run VC diff command and go to the line corresponding to the current."
-  (interactive)
+  (interactive (list current-prefix-arg))
   (with-current-buffer (or (buffer-base-buffer) (current-buffer))
-    (diff-hl-diff-goto-hunk-1)))
+    (diff-hl-diff-goto-hunk-1 historic)))
+
+(defun diff-hl-diff-read-revisions (rev1-default)
+  (let* ((file buffer-file-name)
+         (files (list file))
+         (backend (vc-backend file))
+         (rev2-default nil))
+    (cond
+     ;; if the file is not up-to-date, use working revision as older revision
+     ((not (vc-up-to-date-p file))
+      (setq rev1-default
+            (or rev1-default
+                (vc-working-revision file))))
+     ((not rev1-default)
+      (setq rev1-default (ignore-errors ;If `previous-revision' doesn't work.
+                           (vc-call-backend backend 'previous-revision file
+                                            (vc-working-revision file))))
+      (when (string= rev1-default "") (setq rev1-default nil))))
+    ;; finally read the revisions
+    (let* ((rev1-prompt (if rev1-default
+                            (concat "Older revision (default "
+                                    rev1-default "): ")
+                          "Older revision: "))
+           (rev2-prompt (concat "Newer revision (default "
+                                (or rev2-default "current source") "): "))
+           (rev1 (vc-read-revision rev1-prompt files backend rev1-default))
+           (rev2 (vc-read-revision rev2-prompt files backend rev2-default)))
+      (when (string= rev1 "") (setq rev1 nil))
+      (when (string= rev2 "") (setq rev2 nil))
+      (cons rev1 rev2))))
 
 (defun diff-hl-diff-skip-to (line)
   "In `diff-mode', skip to the hunk and line corresponding to LINE
 in the source file, or the last line of the hunk above it."
+  (goto-char (point-min)) ; Counteract any similar behavior in diff-mode.
   (diff-hunk-next)
   (let (found)
     (while (and (looking-at diff-hunk-header-re-unified) (not found))
@@ -452,6 +549,9 @@ in the source file, or the last line of the hunk above it."
   "Face used to highlight the first column of the hunk to be reverted.")
 
 (defun diff-hl-revert-highlight-first-column (end)
+  (re-search-forward "^[+-]")
+  (forward-line 0)
+  (setq end (diff-hl-split-away-changes 0))
   (let ((inhibit-read-only t))
     (save-excursion
       (while (< (point) end)
@@ -459,58 +559,93 @@ in the source file, or the last line of the hunk above it."
                                          'diff-hl-reverted-hunk-highlight)
         (forward-line 1)))))
 
+(defun diff-hl-revert-narrow-to-hunk (end)
+  (narrow-to-region (point) end))
+
 (defun diff-hl-revert-hunk-1 ()
   (save-restriction
     (widen)
     (vc-buffer-sync)
-    (let ((diff-buffer (generate-new-buffer-name "*diff-hl*"))
-          (buffer (current-buffer))
-          (line (save-excursion
-                  (unless (diff-hl-hunk-overlay-at (point))
-                    (diff-hl-previous-hunk))
-                  (line-number-at-pos)))
-          (fileset (vc-deduce-fileset)))
+    (let* ((diff-buffer (get-buffer-create
+                         (generate-new-buffer-name "*diff-hl*")))
+           (buffer (current-buffer))
+           (line (save-excursion
+                   (diff-hl-find-current-hunk)
+                   (line-number-at-pos)))
+           (file buffer-file-name)
+           (backend (vc-backend file)))
       (unwind-protect
           (progn
-            (vc-diff-internal nil fileset diff-hl-reference-revision nil
-                              nil diff-buffer)
-            (vc-exec-after
-             `(let (beg-line end-line m-end)
+            (vc-setup-buffer diff-buffer)
+            (diff-hl-diff-against-reference file backend diff-buffer)
+            (set-buffer diff-buffer)
+            (diff-mode)
+            (setq-local diff-vc-backend backend)
+            (setq-local diff-vc-revisions (list diff-hl-reference-revision nil))
+            (setq buffer-read-only t)
+            (pop-to-buffer diff-buffer)
+            (vc-run-delayed
+              (vc-diff-finish diff-buffer nil)
+              (let (beg-line end-line m-beg m-end)
                 (when (eobp)
-                  (with-current-buffer ,buffer (diff-hl-remove-overlays))
+                  (with-current-buffer buffer (diff-hl-remove-overlays))
                   (user-error "Buffer is up-to-date"))
-                (let (diff-auto-refine-mode)
-                  (diff-hl-diff-skip-to ,line))
-                (save-excursion
-                  (while (looking-at "[-+]") (forward-line 1))
-                  (setq end-line (line-number-at-pos (point)))
-                  (setq m-end (point-marker))
-                  (unless (eobp) (diff-split-hunk)))
-                (unless (looking-at "[-+]") (forward-line -1))
-                (while (looking-at "[-+]") (forward-line -1))
-                (setq beg-line (line-number-at-pos (point)))
-                (unless (looking-at "@")
-                  (forward-line 1)
-                  (diff-split-hunk))
+                (with-no-warnings
+                  (let (diff-auto-refine-mode)
+                    (diff-hl-diff-skip-to line)))
+                (setq m-end (diff-hl-split-away-changes 3))
+                (setq m-beg (point-marker))
                 (funcall diff-hl-highlight-revert-hunk-function m-end)
+                (setq beg-line (line-number-at-pos m-beg)
+                      end-line (line-number-at-pos m-end))
                 (let ((wbh (window-body-height)))
                   (if (>= wbh (- end-line beg-line))
                       (recenter (/ (+ wbh (- beg-line end-line) 2) 2))
                     (recenter 1)))
-                (when diff-auto-refine-mode
-                  (diff-refine-hunk))
+                (with-no-warnings
+                  (when diff-auto-refine-mode
+                    (diff-refine-hunk)))
                 (if diff-hl-ask-before-revert-hunk
                     (unless (yes-or-no-p (format "Revert current hunk in %s? "
-                                                 ,(cl-caadr fileset)))
+                                                 file))
                       (user-error "Revert canceled")))
                 (let ((diff-advance-after-apply-hunk nil))
                   (save-window-excursion
                     (diff-apply-hunk t)))
-                (with-current-buffer ,buffer
+                (with-current-buffer buffer
                   (save-buffer))
                 (message "Hunk reverted"))))
         (quit-windows-on diff-buffer t)))))
 
+(defun diff-hl-split-away-changes (max-context)
+  "Split away the minimal hunk at point from the rest of the hunk.
+
+The minimal hunk is the hunk a diff program would produce if
+asked for 0 lines of context. Add MAX-CONTEXT lines of context at
+most (stop when encounter another minimal hunk).
+
+Move point to the beginning of the delineated hunk and return
+its end position."
+  (let (end-marker)
+    (save-excursion
+      (while (looking-at "[-+]") (forward-line 1))
+      (dotimes (_i max-context)
+        (unless (looking-at "@\\|[-+]")
+          (forward-line 1)))
+      (setq end-marker (point-marker))
+      (unless (or (eobp)
+                  (looking-at "@"))
+        (diff-split-hunk)))
+    (unless (looking-at "[-+]") (forward-line -1))
+    (while (looking-at "[-+]") (forward-line -1))
+    (dotimes (_i max-context)
+      (unless (looking-at "@\\|[-+]")
+        (forward-line -1)))
+    (unless (looking-at "@")
+      (forward-line 1)
+      (diff-split-hunk))
+    end-marker))
+
 (defun diff-hl-revert-hunk ()
   "Revert the diff hunk with changes at or above the point."
   (interactive)
@@ -522,20 +657,26 @@ in the source file, or the last line of the hunk above it."
            when (overlay-get o 'diff-hl-hunk)
            return o))
 
+(defun diff-hl-search-next-hunk (&optional backward point)
+  "Search the next hunk in the current buffer, or previous if BACKWARD."
+  (save-excursion
+    (when point
+      (goto-char point))
+    (catch 'found
+      (while (not (if backward (bobp) (eobp)))
+        (goto-char (if backward
+                       (previous-overlay-change (point))
+                     (next-overlay-change (point))))
+        (let ((o (diff-hl-hunk-overlay-at (point))))
+          (when (and o (= (overlay-start o) (point)))
+            (throw 'found o)))))))
+
 (defun diff-hl-next-hunk (&optional backward)
   "Go to the beginning of the next hunk in the current buffer."
   (interactive)
-  (let ((pos (save-excursion
-               (catch 'found
-                 (while (not (if backward (bobp) (eobp)))
-                   (goto-char (if backward
-                                  (previous-overlay-change (point))
-                                (next-overlay-change (point))))
-                   (let ((o (diff-hl-hunk-overlay-at (point))))
-                     (when (and o (= (overlay-start o) (point)))
-                       (throw 'found (overlay-start o)))))))))
-    (if pos
-        (goto-char pos)
+  (let ((overlay (diff-hl-search-next-hunk backward)))
+    (if overlay
+        (goto-char (overlay-start overlay))
       (user-error "No further hunks found"))))
 
 (defun diff-hl-previous-hunk ()
@@ -543,6 +684,15 @@ in the source file, or the last line of the hunk above it."
   (interactive)
   (diff-hl-next-hunk t))
 
+(defun diff-hl-find-current-hunk ()
+  (let (o)
+    (cond
+     ((diff-hl-hunk-overlay-at (point)))
+     ((setq o (diff-hl-search-next-hunk t))
+      (goto-char (overlay-start o)))
+     (t
+      (diff-hl-next-hunk)))))
+
 (defun diff-hl-mark-hunk ()
   (interactive)
   (let ((hunk (diff-hl-hunk-overlay-at (point))))
@@ -551,11 +701,84 @@ in the source file, or the last line of the hunk above it."
     (goto-char (overlay-start hunk))
     (push-mark (overlay-end hunk) nil t)))
 
+(defun diff-hl--ensure-staging-supported ()
+  (let ((backend (vc-backend buffer-file-name)))
+    (unless (eq backend 'Git)
+      (user-error "Only Git supports staging; this file is controlled by %s" backend))))
+
+(defun diff-hl-stage-current-hunk ()
+  "Stage the hunk at or near point.
+
+Only supported with Git."
+  (interactive)
+  (diff-hl--ensure-staging-supported)
+  (diff-hl-find-current-hunk)
+  (let* ((line (line-number-at-pos))
+         (file buffer-file-name)
+         (dest-buffer (get-buffer-create " *diff-hl-stage*"))
+         (orig-buffer (current-buffer))
+         (file-base (shell-quote-argument (file-name-nondirectory file)))
+         success)
+    (with-current-buffer dest-buffer
+      (let ((inhibit-read-only t))
+        (erase-buffer)))
+    (diff-hl-diff-buffer-with-reference file dest-buffer nil 3)
+    (with-current-buffer dest-buffer
+      (with-no-warnings
+        (let (diff-auto-refine-mode)
+          (diff-hl-diff-skip-to line)))
+      (let ((inhibit-read-only t))
+        (diff-hl-split-away-changes 3)
+        (save-excursion
+          (diff-end-of-hunk)
+          (delete-region (point) (point-max)))
+        (diff-beginning-of-hunk)
+        (delete-region (point-min) (point))
+        ;; diff-no-select creates a very ugly header; Git rejects it
+        (insert (format "diff a/%s b/%s\n" file-base file-base))
+        (insert (format "--- a/%s\n" file-base))
+        (insert (format "+++ b/%s\n" file-base)))
+      (let ((patchfile (make-temp-file "diff-hl-stage-patch")))
+        (write-region (point-min) (point-max) patchfile
+                      nil 'silent)
+        (unwind-protect
+            (with-current-buffer orig-buffer
+              (with-output-to-string
+                (vc-git-command standard-output 0
+                                patchfile
+                                "apply" "--cached" ))
+              (setq success t))
+          (delete-file patchfile))))
+    (when success
+      (if diff-hl-show-staged-changes
+          (message (concat "Hunk staged; customize `diff-hl-show-staged-changes'"
+                           " to highlight only unstages changes"))
+        (message "Hunk staged"))
+      (unless diff-hl-show-staged-changes
+        (diff-hl-update)))))
+
+(defun diff-hl-unstage-file ()
+  "Unstage all changes in the current file.
+
+Only supported with Git."
+  (interactive)
+  (unless buffer-file-name
+    (user-error "No current file"))
+  (diff-hl--ensure-staging-supported)
+  (vc-git-command nil 0 buffer-file-name "reset")
+  (message "Unstaged all")
+  (unless diff-hl-show-staged-changes
+    (diff-hl-update)))
+
 (defvar diff-hl-command-map
   (let ((map (make-sparse-keymap)))
     (define-key map "n" 'diff-hl-revert-hunk)
     (define-key map "[" 'diff-hl-previous-hunk)
     (define-key map "]" 'diff-hl-next-hunk)
+    (define-key map "*" 'diff-hl-show-hunk)
+    (define-key map "{" 'diff-hl-show-hunk-previous)
+    (define-key map "}" 'diff-hl-show-hunk-next)
+    (define-key map "S" 'diff-hl-stage-current-hunk)
     map))
 (fset 'diff-hl-command-map diff-hl-command-map)
 
@@ -586,7 +809,8 @@ The value of this variable is a mode line template as in
                     ;; saved, in order not to fetch it twice.
                     'find-file-hook)
                   'diff-hl-update-once t t)
-        (add-hook 'vc-checkin-hook 'diff-hl-update nil t)
+        ;; Never removed because it acts globally.
+        (add-hook 'vc-checkin-hook 'diff-hl-after-checkin)
         (add-hook 'after-revert-hook 'diff-hl-after-revert nil t)
         ;; Magit does call `auto-revert-handler', but it usually
         ;; doesn't do much, because `buffer-stale--default-function'
@@ -600,13 +824,25 @@ The value of this variable is a mode line template as in
     (remove-hook 'after-save-hook 'diff-hl-update t)
     (remove-hook 'after-change-functions 'diff-hl-edit t)
     (remove-hook 'find-file-hook 'diff-hl-update t)
-    (remove-hook 'vc-checkin-hook 'diff-hl-update t)
     (remove-hook 'after-revert-hook 'diff-hl-update t)
     (remove-hook 'magit-revert-buffer-hook 'diff-hl-update t)
     (remove-hook 'magit-not-reverted-hook 'diff-hl-update t)
     (remove-hook 'text-scale-mode-hook 'diff-hl-maybe-redefine-bitmaps t)
     (diff-hl-remove-overlays)))
 
+(defun diff-hl-after-checkin ()
+  (let ((fileset (vc-deduce-fileset t)))
+    (dolist (file (nth 1 fileset))
+      (let ((buf (find-buffer-visiting file)))
+        (when buf
+          (with-current-buffer buf
+            (when diff-hl-mode
+              (diff-hl-update))))))))
+
+(defvar diff-hl-repeat-exceptions '(diff-hl-show-hunk
+                                    diff-hl-show-hunk-previous
+                                    diff-hl-show-hunk-next))
+
 (when (require 'smartrep nil t)
   (let (smart-keys)
     (cl-labels ((scan (map)
@@ -614,22 +850,37 @@ The value of this variable is a mode line template as in
                        (lambda (event binding)
                          (if (consp binding)
                              (scan binding)
-                           (when (characterp event)
+                           (when (and (characterp event)
+                                      (not (memq binding diff-hl-repeat-exceptions)))
                              (push (cons (string event) binding) smart-keys))))
                        map)))
       (scan diff-hl-command-map)
       (smartrep-define-key diff-hl-mode-map diff-hl-command-prefix smart-keys))))
 
+;; Integrate with `repeat-mode' in Emacs 28 (https://debbugs.gnu.org/47566)
+;;
+;; While smartrep feels solid, it looks kinda abandoned.  And the
+;; chances of it being put into GNU ELPA are slim too.
+(map-keymap
+ (lambda (_key cmd)
+   (unless (memq cmd diff-hl-repeat-exceptions)
+     (put cmd 'repeat-map 'diff-hl-command-map)))
+ diff-hl-command-map)
+
 (declare-function magit-toplevel "magit-git")
 (declare-function magit-unstaged-files "magit-git")
 
 (defvar diff-hl--magit-unstaged-files nil)
 
 (defun diff-hl-magit-pre-refresh ()
-  (setq diff-hl--magit-unstaged-files (magit-unstaged-files t)))
+  (unless (and diff-hl-disable-on-remote
+               (file-remote-p default-directory))
+   (setq diff-hl--magit-unstaged-files (magit-unstaged-files t))))
 
 (defun diff-hl-magit-post-refresh ()
-  (let* ((topdir (magit-toplevel))
+  (unless (and diff-hl-disable-on-remote
+               (file-remote-p default-directory))
+   (let* ((topdir (magit-toplevel))
          (modified-files
           (mapcar (lambda (file) (expand-file-name file topdir))
                   (delete-consecutive-dups
@@ -659,7 +910,7 @@ The value of this variable is a mode line template as in
                 (diff-hl-update))
                ((not (memq (vc-state file backend) unmodified-states))
                 (vc-state-refresh file backend)
-                (diff-hl-update))))))))))
+                (diff-hl-update)))))))))))
 
 (defun diff-hl-dir-update ()
   (dolist (pair (if (vc-dir-marked-files)
@@ -678,12 +929,123 @@ The value of this variable is a mode line template as in
       (add-hook 'vc-checkin-hook 'diff-hl-dir-update t t)
     (remove-hook 'vc-checkin-hook 'diff-hl-dir-update t)))
 
+(defun diff-hl-make-temp-file-name (file rev &optional manual)
+  "Return a backup file name for REV or the current version of FILE.
+If MANUAL is non-nil it means that a name for backups created by
+the user should be returned."
+  (let* ((auto-save-file-name-transforms
+          `((".*" ,temporary-file-directory t)))
+         (buffer-file-name file))
+    (expand-file-name
+     (concat (make-auto-save-file-name)
+             ".~" (subst-char-in-string
+                   ?/ ?_ rev)
+             (unless manual ".") "~")
+     temporary-file-directory)))
+
+(defun diff-hl-create-revision (file revision)
+  "Read REVISION of FILE into a buffer and return the buffer."
+  (let ((automatic-backup (diff-hl-make-temp-file-name file revision))
+        (filebuf (get-file-buffer file))
+        (filename (diff-hl-make-temp-file-name file revision 'manual)))
+    (unless (file-exists-p filename)
+      (if (file-exists-p automatic-backup)
+          (rename-file automatic-backup filename nil)
+        (with-current-buffer filebuf
+          (let ((coding-system-for-read 'no-conversion)
+                (coding-system-for-write 'no-conversion))
+            (condition-case nil
+                (with-temp-file filename
+                  (let ((outbuf (current-buffer)))
+                    ;; Change buffer to get local value of
+                    ;; vc-checkout-switches.
+                    (with-current-buffer filebuf
+                      (vc-call find-revision file revision outbuf))))
+              (error
+               (when (file-exists-p filename)
+                 (delete-file filename))))))))
+    filename))
+
+(defun diff-hl-working-revision (file &optional backend)
+  "Like vc-working-revision, but always up-to-date"
+  (vc-file-setprop file 'vc-working-revision
+                   (vc-call-backend (or backend (vc-backend file))
+                                    'working-revision file)))
+
+(declare-function diff-no-select "diff")
+
+(defun diff-hl-diff-buffer-with-reference (file &optional dest-buffer backend context-lines)
+  "Compute the diff between the current buffer contents and reference in BACKEND.
+The diffs are computed in the buffer DEST-BUFFER. This requires
+the `diff-program' to be in your `exec-path'.
+CONTEXT-LINES is the size of the unified diff context, defaults to 0."
+  (require 'diff)
+  (vc-ensure-vc-buffer)
+  (save-current-buffer
+    (let* ((dest-buffer (or dest-buffer "*diff-hl-diff-buffer-with-reference*"))
+           (backend (or backend (vc-backend file)))
+           (temporary-file-directory
+            (if (file-directory-p "/dev/shm/")
+                "/dev/shm/"
+              temporary-file-directory))
+           (rev
+            (if (and (eq backend 'Git)
+                     (not diff-hl-reference-revision)
+                     (not diff-hl-show-staged-changes))
+                (diff-hl-git-index-revision
+                 file
+                 (diff-hl-git-index-object-name file))
+              (diff-hl-create-revision
+               file
+               (or diff-hl-reference-revision
+                   (diff-hl-working-revision file backend)))))
+           (switches (format "-U %d --strip-trailing-cr" (or context-lines 0))))
+      (diff-no-select rev (current-buffer) switches 'noasync
+                      (get-buffer-create dest-buffer))
+      (with-current-buffer dest-buffer
+        (let ((inhibit-read-only t))
+          ;; Function `diff-sentinel' adds a final line, so remove it
+          (delete-matching-lines "^Diff finished.*")))
+      (get-buffer-create dest-buffer))))
+
+;; TODO: Cache based on .git/index's mtime, maybe.
+(defun diff-hl-git-index-object-name (file)
+  (with-temp-buffer
+    (vc-git-command (current-buffer) 0 file "ls-files" "-s")
+    (and
+     (goto-char (point-min))
+     (re-search-forward "^[0-9]+ \\([0-9a-f]+\\)")
+     (match-string-no-properties 1))))
+
+(defun diff-hl-git-index-revision (file object-name)
+  (let ((filename (diff-hl-make-temp-file-name file
+                                               (concat ":" object-name)
+                                               'manual))
+        (filebuf (get-file-buffer file)))
+    (unless (file-exists-p filename)
+      (with-current-buffer filebuf
+        (let ((coding-system-for-read 'no-conversion)
+              (coding-system-for-write 'no-conversion))
+          (condition-case nil
+              (with-temp-file filename
+                (let ((outbuf (current-buffer)))
+                  ;; Change buffer to be inside the repo.
+                  (with-current-buffer filebuf
+                    (vc-git-command outbuf 0 nil
+                                    "cat-file" "blob" object-name))))
+            (error
+             (when (file-exists-p filename)
+               (delete-file filename)))))))
+    filename))
+
 ;;;###autoload
 (defun turn-on-diff-hl-mode ()
   "Turn on `diff-hl-mode' or `diff-hl-dir-mode' in a buffer if appropriate."
   (cond
    (buffer-file-name
-    (diff-hl-mode 1))
+    (unless (and diff-hl-disable-on-remote
+                 (file-remote-p buffer-file-name))
+      (diff-hl-mode 1)))
    ((eq major-mode 'vc-dir-mode)
     (diff-hl-dir-mode 1))))
 
@@ -697,6 +1059,59 @@ The value of this variable is a mode line template as in
               (t (memq major-mode diff-hl-global-modes)))
     (turn-on-diff-hl-mode)))
 
+(declare-function vc-annotate-extract-revision-at-line "vc-annotate")
+(declare-function diff-hl-amend-mode "diff-hl-amend")
+
+;;;###autoload
+(defun diff-hl-set-reference-rev (rev)
+  "Set the reference revision globally to REV.
+When called interactively, REV read with completion.
+
+The default value chosen using one of methods below:
+
+- In a log view buffer, it uses the revision of current entry.
+Call `vc-print-log' or `vc-print-root-log' first to open a log
+view buffer.
+- In a VC annotate buffer, it uses the revision of current line.
+- In other situations, it uses the symbol at point.
+
+Notice that this sets the reference revision globally, so in
+files from other repositories, `diff-hl-mode' will not highlight
+changes correctly, until you run `diff-hl-reset-reference-rev'.
+
+Also notice that this will disable `diff-hl-amend-mode' in
+buffers that enables it, since `diff-hl-amend-mode' overrides its
+effect."
+  (interactive
+   (let* ((def (or (and (equal major-mode 'vc-annotate-mode)
+                        (car (vc-annotate-extract-revision-at-line)))
+                   (log-view-current-tag)
+                   (thing-at-point 'symbol t)))
+          (prompt (if def
+                      (format "Reference revision (default %s): " def)
+                    "Reference revision: ")))
+     (list (vc-read-revision prompt nil nil def))))
+  (if rev
+      (message "Set reference revision to %s" rev)
+    (user-error "No reference revision specified"))
+  (setq diff-hl-reference-revision rev)
+  (dolist (buf (buffer-list))
+    (with-current-buffer buf
+      (when diff-hl-mode
+        (when (bound-and-true-p diff-hl-amend-mode)
+          (diff-hl-amend-mode -1))
+        (diff-hl-update)))))
+
+;;;###autoload
+(defun diff-hl-reset-reference-rev ()
+  "Reset the reference revision globally to the most recent one."
+  (interactive)
+  (setq diff-hl-reference-revision nil)
+  (dolist (buf (buffer-list))
+    (with-current-buffer buf
+      (when diff-hl-mode
+        (diff-hl-update)))))
+
 ;;;###autoload
 (define-globalized-minor-mode global-diff-hl-mode diff-hl-mode
   diff-hl--global-turn-on :after-hook (diff-hl-global-mode-change))
